from datetime import datetime
from django.db import models
from django.conf import settings
from django.contrib.auth.models import UserManager


MEDIAWIKI_DB_ALIAS = getattr(settings, 'MEDIAWIKI_DB_ALIAS', 'mediawiki')
MEDIAWIKI_TABLE_PREFIX = getattr(settings, 'MEDIAWIKI_TABLE_PREFIX', '')
MEDIAWIKI_USER_FIELD = getattr(settings, 'MEDIAWIKI_USER_FIELD', 'mediawiki_user')
MEDIAWIKI_USER_FK_FIELD = getattr(settings, 'MEDIAWIKI_USER_FK_FIELD', 'mediawiki_user_id')


def ts_to_dt(ts):
    if len(ts) == 14:
        return datetime(
            int(ts[:4]),
            int(ts[4:6]),
            int(ts[6:8]),
            int(ts[8:10]),
            int(ts[10:12]),
            int(ts[12:]),
        )


class MediaWikiEnabledUserManager(UserManager):

    def create_from_mediawiki_user(self, wiki_user):
        email = self.normalize_email(wiki_user.email)
        username = self.model.normalize_username(wiki_user.name)
        first, last = self.normalize_first_last(wiki_user)
        user = self.model(
            username=username,
            email=email,
            first_name=first,
            last_name=last,
            date_joined=wiki_user.registered
        )
        setattr(user, MEDIAWIKI_TABLE_PREFIX, wiki_user)
        setattr(user, MEDIAWIKI_USER_FK_FIELD, wiki_user.id)
        user.save(using=self._db)
        return user

    @staticmethod
    def normalize_first_last(wiki_user):
        first, last = '', ''
        if wiki_user.real_name:
            parts = wiki_user.real_name.split(' ')
            first = parts[0]
            if len(parts) > 1:
                last = ' '.join(parts[1:])
        return first, last


class MediaWikiUserManager(models.Manager):

    def get_queryset(self):
        return super().get_queryset().using(MEDIAWIKI_DB_ALIAS)


class MediaWikiUser(models.Model):

    class Meta:
        verbose_name = "MediaWiki User"
        verbose_name_plural = "MediaWiki Users"
        ordering = ('name',)
        db_table = MEDIAWIKI_TABLE_PREFIX+'user'
        managed = False

    id = models.AutoField(primary_key=True, db_column='user_id')
    name = models.BinaryField(max_length=255, db_column='user_name')
    real_name = models.CharField(max_length=255, db_column='user_real_name')
    email = models.CharField(max_length=255, db_column='user_email')

    token = models.BinaryField(max_length=32, db_column='user_token')

    email_authenticated = models.BinaryField(max_length=14, db_column='user_email_authenticated')
    email_token = models.BinaryField(max_length=32, db_column='user_email_token')
    email_token_expires = models.BinaryField(max_length=14, db_column='user_email_token_expires')

    registration = models.BinaryField(max_length=14, db_column='user_registration')
    touched = models.BinaryField(max_length=14, db_column='user_touched')

    edit_count = models.PositiveIntegerField(db_column='user_editcount')

    objects = MediaWikiUserManager()

    def verify_session_and_cookie_values(self, session, user_id):
        return self.id == session[b'wsUserID'] == int(user_id)

    @property
    def registered(self):
        return ts_to_dt(self.registration)

    @property
    def last_accessed(self):
        return ts_to_dt(self.touched)


class RecentChangesManager(models.Manager):

    def get_queryset(self):
        return super().get_queryset().using(MEDIAWIKI_DB_ALIAS)


class RecentChanges(models.Model):

    class Meta:
        db_table = MEDIAWIKI_TABLE_PREFIX+'recentchanges'
        managed = False
        ordering = ('-rc_timestamp',)

    id = models.AutoField(primary_key=True, db_column='rc_id')
    title = models.CharField(max_length=255, db_column='rc_title')
    user = models.ForeignKey(MediaWikiUser, db_column='rc_user')
    username = models.CharField(max_length=255, db_column='rc_user_text')
    rc_timestamp = models.BinaryField()

    objects = RecentChangesManager()

    @property
    def modified(self):
        return ts_to_dt(self.rc_timestamp)


class RevisionManager(models.Manager):

    def get_queryset(self):
        return super().get_queryset().using(MEDIAWIKI_DB_ALIAS)


class Revision(models.Model):

    class Meta:
        db_table = MEDIAWIKI_TABLE_PREFIX+'revision'
        managed = False

    id = models.AutoField(primary_key=True, db_column='rev_id')
    rev_timestamp = models.BinaryField()

    objects = RevisionManager()

    @property
    def modified(self):
        return ts_to_dt(self.rev_timestamp)



import re
import zlib
import phpserialize
from django.db import connections
from django.conf import settings
from django.contrib.auth import get_user as get_django_user_from_request
from django.contrib.auth import get_user_model as get_django_user_model
from django.contrib.auth import login

from .models import (
    MediaWikiUser,
    MEDIAWIKI_USER_FK_FIELD,
    MEDIAWIKI_USER_FIELD,
    MEDIAWIKI_TABLE_PREFIX,
    MEDIAWIKI_DB_ALIAS
)

MEDIAWIKI_COOKIE_PREFIX = getattr(settings, 'MEDIAWIKI_COOKIE_PREFIX', 'mediawiki')


def unserialize_session(val):
    session = {}
    groups = re.split(b"([a-zA-Z0-9_]+)\|", val)
    if len(groups) > 2:
        groups = groups[1:]
        groups = zip(*([iter(groups)] * 2))
    for key, php_value in groups:
        session[key] = phpserialize.loads(php_value)
    return session


def get_session(session_id):
    with connections[MEDIAWIKI_DB_ALIAS].cursor() as c:
        c.execute("select value from {}objectcache where keyname = %s".format(MEDIAWIKI_TABLE_PREFIX), (
            'mediawiki:session:'+session_id,
        ))
        result = c.fetchone()
        if result and result[0]:
            serialized = zlib.decompress(result[0], -15)
            return unserialize_session(serialized)


def get_session_from_request(request):
    session_id = request.COOKIES.get(MEDIAWIKI_COOKIE_PREFIX+'_session')
    if session_id:
        return get_session(session_id)


def get_user(user_id):
    try:
        return MediaWikiUser.objects.get(pk=user_id)
    except (MediaWikiUser.DoesNotExist, MediaWikiUser.MultipleObjectsReturned):
        return None


def get_mediawiki_user_from_request(request):
    session = get_session_from_request(request)
    if session and session.get(b'wsUserID'):
        user = get_user(session[b'wsUserID'])
        if user:
            user_id = request.COOKIES.get(MEDIAWIKI_COOKIE_PREFIX+'UserID')
            if user.verify_session_and_cookie_values(session, user_id):
                return user


def get_django_user_from_mediawiki_user(wiki_user):
    User = get_django_user_model()
    try:
        return User.objects.get(**{MEDIAWIKI_USER_FK_FIELD: wiki_user.id})
    except (User.DoesNotExist, User.MultipleObjectsReturned):
        return None


def get_or_create_django_user_from_mediawiki_user(wiki_user):
    DjangoUser = get_django_user_model()
    try:
        return DjangoUser.objects.get(**{MEDIAWIKI_USER_FK_FIELD: wiki_user.id}), False
    except DjangoUser.DoesNotExist:
        return DjangoUser.objects.create_from_mediawiki_user(wiki_user), True


def get_or_create_django_user(request):
    from django.contrib.auth.models import AnonymousUser

    wiki_user = get_mediawiki_user_from_request(request)
    if wiki_user:

        django_user = get_django_user_from_request(request)
        if django_user.is_authenticated:

            if wiki_user.id == getattr(django_user, MEDIAWIKI_USER_FK_FIELD):
                # wiki session and django sessions match
                setattr(django_user, MEDIAWIKI_USER_FIELD, wiki_user)
                return django_user
            else:
                # sessions mismatch, invalidate django session
                request.session.flush()

        else:

            django_user, _ = get_or_create_django_user_from_mediawiki_user(wiki_user)
            login(request, django_user)
            setattr(django_user, MEDIAWIKI_USER_FIELD, wiki_user)
            return django_user

    return AnonymousUser()



from django.contrib import admin
from .models import MediaWikiUser
from .mediawiki import (
    get_django_user_from_mediawiki_user,
    get_or_create_django_user_from_mediawiki_user
)


@admin.register(MediaWikiUser)
class MediaWikiUserAdmin(admin.ModelAdmin):
    list_per_page = 20
    list_display = (
        'name', 'real_name', 'email', 'last_accessed', 'registered', 'edit_count', 'linked'
    )
    search_fields = (
        'name', 'real_name', 'email'
    )
    fields = readonly_fields = ('id',)+list_display

    def get_actions(self, request):
        return {'link': (MediaWikiUserAdmin.link, 'link', "Link selected MediaWiki users into django")}

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def link(self, request, queryset):
        users = []
        for user in queryset:
            _, created = get_or_create_django_user_from_mediawiki_user(user)
            if created:
                users.append("'{}'".format(user.name))
        self.message_user(request, "%s successfully linked.".format(', '.join(users)))

    def linked(self, obj):
        return bool(get_django_user_from_mediawiki_user(obj))
    linked.boolean = True
    linked.short_description = 'Linked'



from django.conf import settings
from django.utils.deprecation import MiddlewareMixin
from django.utils.functional import SimpleLazyObject
from mediawiki_auth import mediawiki


def get_user(request):
    if not hasattr(request, '_cached_user'):
        request._cached_user = mediawiki.get_or_create_django_user(request)
    return request._cached_user


class AuthenticationMiddleware(MiddlewareMixin):
    def process_request(self, request):
        assert hasattr(request, 'session'), (
            "The Django authentication middleware requires session middleware "
            "to be installed. Edit your MIDDLEWARE%s setting to insert "
            "'django.contrib.sessions.middleware.SessionMiddleware' before "
            "'django.contrib.auth.middleware.AuthenticationMiddleware'."
        ) % ("_CLASSES" if settings.MIDDLEWARE is None else "")
        request.user = SimpleLazyObject(lambda: get_user(request))
